"""
Prometheus metrics exported by Jupyter Notebook Server

Read https://prometheus.io/docs/practices/naming/ for naming
conventions for metrics & labels.
"""


from prometheus_client import Histogram, Gauge


HTTP_REQUEST_DURATION_SECONDS = Histogram(
    'http_request_duration_seconds',
    'duration in seconds for all HTTP requests',
    ['method', 'handler', 'status_code'],
)

TERMINAL_CURRENTLY_RUNNING_TOTAL = Gauge(
    'terminal_currently_running_total',
    'counter for how many terminals are running',
)

KERNEL_CURRENTLY_RUNNING_TOTAL = Gauge(
    'kernel_currently_running_total',
    'counter for how many kernels are running labeled by type',
    ['type']
)



INITIAL_CELLS = ['print("a")', 'print("b")', 'print("c")']
def test_multiselect_toggle(prefill_notebook):
    notebook = prefill_notebook(INITIAL_CELLS)
    def extend_selection_by(delta):
        notebook.browser.execute_script(
            "Jupyter.notebook.extend_selection_by(arguments[0]);", delta)

    def n_selected_cells():
        return notebook.browser.execute_script(
            "return Jupyter.notebook.get_selected_cells().length;")

    def select_cells():
        notebook.focus_cell(0)
        extend_selection_by(2)

    # Test that cells, which start off not collapsed, are collapsed after
    # calling the multiselected cell toggle.
    select_cells()
    assert n_selected_cells() == 3
    notebook.browser.execute_script("Jupyter.notebook.execute_selected_cells();")
    select_cells()
    notebook.browser.execute_script("Jupyter.notebook.toggle_cells_outputs();")
    cell_output_states = notebook.browser.execute_script(
        "return Jupyter.notebook.get_cells().map(c => c.collapsed)")
    assert cell_output_states == [False] * 3, "ensure that all cells are not collapsed"

    # Test that cells, which start off not scrolled are scrolled after
    # calling the multiselected scroll toggle.
    select_cells()
    assert n_selected_cells() == 3
    notebook.browser.execute_script("Jupyter.notebook.toggle_cells_outputs_scroll();")
    cell_scrolled_states = notebook.browser.execute_script(
        "return Jupyter.notebook.get_cells().map(c => c.output_area.scroll_state)")
    assert all(cell_scrolled_states), "ensure that all have scrolling enabled"

    # Test that cells, which start off not cleared are cleared after
    # calling the multiselected scroll toggle.
    select_cells()
    assert n_selected_cells() == 3
    notebook.browser.execute_script("Jupyter.notebook.clear_cells_outputs();")
    cell_outputs_cleared = notebook.browser.execute_script(
        "return Jupyter.notebook.get_cells().map(c => c.output_area.element.html())")
    assert cell_outputs_cleared == [""] * 3, "ensure that all cells are cleared"



from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from .utils import shift, cmdtrl


def test_execute_code(notebook):
    browser = notebook.browser
    
    def clear_outputs():
        return notebook.browser.execute_script(
            "Jupyter.notebook.clear_all_output();")

    # Execute cell with Javascript API
    notebook.edit_cell(index=0, content='a=10; print(a)')
    browser.execute_script("Jupyter.notebook.get_cell(0).execute();")
    outputs = notebook.wait_for_cell_output(0)
    assert outputs[0].text == '10'

    # Execute cell with Shift-Enter
    notebook.edit_cell(index=0, content='a=11; print(a)')
    clear_outputs()
    shift(notebook.browser, Keys.ENTER)
    outputs = notebook.wait_for_cell_output(0)
    assert outputs[0].text == '11'
    notebook.delete_cell(index=1)

    # Execute cell with Ctrl-Enter
    notebook.edit_cell(index=0, content='a=12; print(a)')
    clear_outputs()
    cmdtrl(notebook.browser, Keys.ENTER)
    outputs = notebook.wait_for_cell_output(0)
    assert outputs[0].text == '12'

    # Execute cell with toolbar button
    notebook.edit_cell(index=0, content='a=13; print(a)')
    clear_outputs()
    notebook.browser.find_element(
        By.CSS_SELECTOR,
        "button[data-jupyter-action='jupyter-notebook:run-cell-and-select-next']").click()
    outputs = notebook.wait_for_cell_output(0)
    assert outputs[0].text == '13'

    # Set up two cells to test stopping on error
    notebook.edit_cell(index=0, content='raise IOError')
    notebook.edit_cell(index=1, content='a=14; print(a)')

    # Default behaviour: stop on error
    clear_outputs()
    browser.execute_script("""
        var cell0 = Jupyter.notebook.get_cell(0);
        var cell1 = Jupyter.notebook.get_cell(1);
        cell0.execute();
        cell1.execute();
    """)
    outputs = notebook.wait_for_cell_output(0)
    assert notebook.get_cell_output(1) == []

    # Execute a cell with stop_on_error=false
    clear_outputs()
    browser.execute_script("""
            var cell0 = Jupyter.notebook.get_cell(0);
            var cell1 = Jupyter.notebook.get_cell(1);
            cell0.execute(false);
            cell1.execute();
        """)
    outputs = notebook.wait_for_cell_output(1)
    assert outputs[0].text == '14'




"""Tornado handlers for the contents web service.

Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-27%3A-Contents-Service
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json

from tornado import gen, web

from notebook.utils import maybe_future, url_path_join, url_escape
try:
    from jupyter_client.jsonutil import json_default
except ImportError:
    from jupyter_client.jsonutil import (
        date_default as json_default
    )

from notebook.base.handlers import (
    IPythonHandler, APIHandler, path_regex,
)


def validate_model(model, expect_content):
    """
    Validate a model returned by a ContentsManager method.

    If expect_content is True, then we expect non-null entries for 'content'
    and 'format'.
    """
    required_keys = {
        "name",
        "path",
        "type",
        "writable",
        "created",
        "last_modified",
        "mimetype",
        "content",
        "format",
    }
    missing = required_keys - set(model.keys())
    if missing:
        raise web.HTTPError(
            500,
            f"Missing Model Keys: {missing}",
        )

    maybe_none_keys = ['content', 'format']
    if expect_content:
        errors = [key for key in maybe_none_keys if model[key] is None]
        if errors:
            raise web.HTTPError(
                500,
                f"Keys unexpectedly None: {errors}",
            )
    else:
        errors = {
            key: model[key]
            for key in maybe_none_keys
            if model[key] is not None
        }
        if errors:
            raise web.HTTPError(
                500,
                f"Keys unexpectedly not None: {errors}",
            )


class ContentsHandler(APIHandler):

    def location_url(self, path):
        """Return the full URL location of a file.

        Parameters
        ----------
        path : unicode
            The API path of the file, such as "foo/bar.txt".
        """
        return url_path_join(
            self.base_url, 'api', 'contents', url_escape(path)
        )

    def _finish_model(self, model, location=True):
        """Finish a JSON request with a model, setting relevant headers, etc."""
        if location:
            location = self.location_url(model['path'])
            self.set_header('Location', location)
        self.set_header('Last-Modified', model['last_modified'])
        self.set_header('Content-Type', 'application/json')
        self.finish(json.dumps(model, default=json_default))

    @web.authenticated
    @gen.coroutine
    def get(self, path=''):
        """Return a model for a file or directory.

        A directory model contains a list of models (without content)
        of the files and directories it contains.
        """
        path = path or ''
        cm = self.contents_manager
        type = self.get_query_argument('type', default=None)
        if type not in {None, 'directory', 'file', 'notebook'}:
            raise web.HTTPError(400, f'Type {type!r} is invalid')

        format = self.get_query_argument('format', default=None)
        if format not in {None, 'text', 'base64'}:
            raise web.HTTPError(400, f'Format {format!r} is invalid')
        content = self.get_query_argument('content', default='1')
        if content not in {'0', '1'}:
            raise web.HTTPError(400, f'Content {content!r} is invalid')
        content = int(content)
        if cm.is_hidden(path) and not cm.allow_hidden:
            raise web.HTTPError(404, f'file or directory {path!r} does not exist')
        model = yield maybe_future(self.contents_manager.get(
            path=path, type=type, format=format, content=content,
        ))
        validate_model(model, expect_content=content)
        self._finish_model(model, location=False)

    @web.authenticated
    @gen.coroutine
    def patch(self, path=''):
        """PATCH renames a file or directory without re-uploading content."""
        cm = self.contents_manager
        model = self.get_json_body()
        old_path = model.get('path')
        if old_path and (cm.is_hidden(path) or cm.is_hidden(old_path))  and not cm.allow_hidden:
            raise web.HTTPError(400, f'Cannot rename file or directory {path!r}')
        if model is None:
            raise web.HTTPError(400, 'JSON body missing')
        model = yield maybe_future(cm.update(model, path))
        validate_model(model, expect_content=False)
        self._finish_model(model)

    @gen.coroutine
    def _copy(self, copy_from, copy_to=None):
        """Copy a file, optionally specifying a target directory."""
        self.log.info(f"Copying {copy_from} to {copy_to or ''}")
        model = yield maybe_future(self.contents_manager.copy(copy_from, copy_to))
        self.set_status(201)
        validate_model(model, expect_content=False)
        self._finish_model(model)

    @gen.coroutine
    def _upload(self, model, path):
        """Handle upload of a new file to path"""
        self.log.info("Uploading file to %s", path)
        model = yield maybe_future(self.contents_manager.new(model, path))
        self.set_status(201)
        validate_model(model, expect_content=False)
        self._finish_model(model)

    @gen.coroutine
    def _new_untitled(self, path, type='', ext=''):
        """Create a new, empty untitled entity"""
        self.log.info("Creating new %s in %s", type or 'file', path)
        model = yield maybe_future(self.contents_manager.new_untitled(path=path, type=type, ext=ext))
        self.set_status(201)
        validate_model(model, expect_content=False)
        self._finish_model(model)

    @gen.coroutine
    def _save(self, model, path):
        """Save an existing file."""
        chunk = model.get("chunk", None)
        if not chunk or chunk == -1:  # Avoid tedious log information
            self.log.info("Saving file at %s", path)
        model = yield maybe_future(self.contents_manager.save(model, path))
        validate_model(model, expect_content=False)
        self._finish_model(model)

    @web.authenticated
    @gen.coroutine
    def post(self, path=''):
        """Create a new file in the specified path.

        POST creates new files. The server always decides on the name.

        POST /api/contents/path
          New untitled, empty file or directory.
        POST /api/contents/path
          with body {"copy_from" : "/path/to/OtherNotebook.ipynb"}
          New copy of OtherNotebook in path
        """

        cm = self.contents_manager

        file_exists = yield maybe_future(cm.file_exists(path))
        if file_exists:
            raise web.HTTPError(400, "Cannot POST to files, use PUT instead.")

        dir_exists = yield maybe_future(cm.dir_exists(path))
        if not dir_exists:
            raise web.HTTPError(404, f"No such directory: {path}")

        model = self.get_json_body()
        copy_from = model.get('copy_from')
        if copy_from and (cm.is_hidden(path) or cm.is_hidden(copy_from))  and not cm.allow_hidden:
            raise web.HTTPError(400, f'Cannot copy file or directory {path!r}')

        if model is not None:
            copy_from = model.get('copy_from')
            ext = model.get('ext', '')
            type = model.get('type', '')
            if copy_from:
                yield self._copy(copy_from, path)
            else:
                yield self._new_untitled(path, type=type, ext=ext)
        else:
            yield self._new_untitled(path)

    @web.authenticated
    @gen.coroutine
    def put(self, path=''):
        """Saves the file in the location specified by name and path.

        PUT is very similar to POST, but the requester specifies the name,
        whereas with POST, the server picks the name.

        PUT /api/contents/path/Name.ipynb
          Save notebook at ``path/Name.ipynb``. Notebook structure is specified
          in `content` key of JSON request body. If content is not specified,
          create a new empty notebook.
        """
        model = self.get_json_body()
        cm = self.contents_manager
        if model:
            if model.get('copy_from'):
                raise web.HTTPError(400, "Cannot copy with PUT, only POST")
            if model.get('path') and (cm.is_hidden(path) or cm.is_hidden(model.get('path')))  and not cm.allow_hidden:
                raise web.HTTPError(400, f'Cannot create file or directory {path!r}')
            exists = yield maybe_future(self.contents_manager.file_exists(path))
            if exists:
                yield maybe_future(self._save(model, path))
            else:
                yield maybe_future(self._upload(model, path))
        else:
            yield maybe_future(self._new_untitled(path))

    @web.authenticated
    @gen.coroutine
    def delete(self, path=''):
        """delete a file in the given path"""
        cm = self.contents_manager

        if cm.is_hidden(path) and not cm.allow_hidden:
            raise web.HTTPError(400, f'Cannot delete file or directory {path!r}')

        self.log.warning('delete %s', path)
        yield maybe_future(cm.delete(path))
        self.set_status(204)
        self.finish()


class CheckpointsHandler(APIHandler):

    @web.authenticated
    @gen.coroutine
    def get(self, path=''):
        """get lists checkpoints for a file"""
        cm = self.contents_manager
        checkpoints = yield maybe_future(cm.list_checkpoints(path))
        data = json.dumps(checkpoints, default=json_default)
        self.finish(data)

    @web.authenticated
    @gen.coroutine
    def post(self, path=''):
        """post creates a new checkpoint"""
        cm = self.contents_manager
        checkpoint = yield maybe_future(cm.create_checkpoint(path))
        data = json.dumps(checkpoint, default=json_default)
        location = url_path_join(self.base_url, 'api/contents',
            url_escape(path), 'checkpoints', url_escape(checkpoint['id']))
        self.set_header('Location', location)
        self.set_status(201)
        self.finish(data)


class ModifyCheckpointsHandler(APIHandler):

    @web.authenticated
    @gen.coroutine
    def post(self, path, checkpoint_id):
        """post restores a file from a checkpoint"""
        cm = self.contents_manager
        yield maybe_future(cm.restore_checkpoint(checkpoint_id, path))
        self.set_status(204)
        self.finish()

    @web.authenticated
    @gen.coroutine
    def delete(self, path, checkpoint_id):
        """delete clears a checkpoint for a given file"""
        cm = self.contents_manager
        yield maybe_future(cm.delete_checkpoint(checkpoint_id, path))
        self.set_status(204)
        self.finish()


class NotebooksRedirectHandler(IPythonHandler):
    """Redirect /api/notebooks to /api/contents"""
    SUPPORTED_METHODS = ('GET', 'PUT', 'PATCH', 'POST', 'DELETE')

    def get(self, path):
        self.log.warning("/api/notebooks is deprecated, use /api/contents")
        self.redirect(url_path_join(
            self.base_url,
            'api/contents',
            path
        ))

    put = patch = post = delete = get


class TrustNotebooksHandler(IPythonHandler):
    """ Handles trust/signing of notebooks """

    @web.authenticated
    @gen.coroutine
    def post(self,path=''):
        cm = self.contents_manager
        yield maybe_future(cm.trust_notebook(path))
        self.set_status(201)
        self.finish()
#-----------------------------------------------------------------------------
# URL to handler mappings
#-----------------------------------------------------------------------------


_checkpoint_id_regex = r"(?P<checkpoint_id>[\w-]+)"

default_handlers = [
    (fr"/api/contents{path_regex}/checkpoints", CheckpointsHandler),
    (fr"/api/contents{path_regex}/checkpoints/{_checkpoint_id_regex}",
        ModifyCheckpointsHandler),
    (fr"/api/contents{path_regex}/trust", TrustNotebooksHandler),
    (fr"/api/contents{path_regex}", ContentsHandler),
    (r"/api/notebooks/?(.*)", NotebooksRedirectHandler),
]



"""Tests for file IO"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os.path
import unittest
import pytest
import stat
import sys

from ..fileio import atomic_writing

from ipython_genutils.tempdir import TemporaryDirectory

umask = 0

def test_atomic_writing():
    class CustomExc(Exception):
        pass

    with TemporaryDirectory() as td:
        f1 = os.path.join(td, 'penguin')
        with open(f1, 'w') as f:
            f.write('Before')

        if os.name != 'nt':
            os.chmod(f1, 0o701)
            orig_mode = stat.S_IMODE(os.stat(f1).st_mode)

        f2 = os.path.join(td, 'flamingo')
        try:
            os.symlink(f1, f2)
            have_symlink = True
        except (AttributeError, NotImplementedError, OSError):
            # AttributeError: Python doesn't support it
            # NotImplementedError: The system doesn't support it
            # OSError: The user lacks the privilege (Windows)
            have_symlink = False

        with pytest.raises(CustomExc):
            with atomic_writing(f1) as f:
                f.write('Failing write')
                raise CustomExc

        # Because of the exception, the file should not have been modified
        with open(f1) as f:
            assert f.read() == 'Before'

        with atomic_writing(f1) as f:
            f.write('Overwritten')

        with open(f1) as f:
            assert f.read() == 'Overwritten'

        if os.name != 'nt':
            mode = stat.S_IMODE(os.stat(f1).st_mode)
            assert mode == orig_mode

        if have_symlink:
            # Check that writing over a file preserves a symlink
            with atomic_writing(f2) as f:
                f.write('written from symlink')

            with open(f1) as f:
                assert f.read() == 'written from symlink'

class TestWithSetUmask(unittest.TestCase):
    def setUp(self):
        # save umask
        global umask
        umask = os.umask(0)
        os.umask(umask)

    def tearDown(self):
        # restore umask
        os.umask(umask)

    @pytest.mark.skipif(sys.platform == "win32", reason="do not run on windows")
    def test_atomic_writing_umask(self):
        with TemporaryDirectory() as td:
            os.umask(0o022)
            f1 = os.path.join(td, '1')
            with atomic_writing(f1) as f:
                f.write('1')
            mode = stat.S_IMODE(os.stat(f1).st_mode)
            assert mode == 0o644

            os.umask(0o057)
            f2 = os.path.join(td, '2')
            with atomic_writing(f2) as f:
                f.write('2')
            mode = stat.S_IMODE(os.stat(f2).st_mode)
            assert mode == 0o620


def test_atomic_writing_newlines():
    with TemporaryDirectory() as td:
        path = os.path.join(td, 'testfile')

        lf = 'a\nb\nc\n'
        plat = lf.replace('\n', os.linesep)
        crlf = lf.replace('\n', '\r\n')

        # test default
        with open(path, 'w') as f:
            f.write(lf)
        with open(path, newline='') as f:
            read = f.read()
        assert read == plat

        # test newline=LF
        with open(path, 'w', newline='\n') as f:
            f.write(lf)
        with open(path, newline='') as f:
            read = f.read()
        assert read == lf

        # test newline=CRLF
        with atomic_writing(path, newline='\r\n') as f:
            f.write(lf)
        with open(path, newline='') as f:
            read = f.read()
        assert read == crlf

        # test newline=no convert
        text = 'crlf\r\ncr\rlf\n'
        with atomic_writing(path, newline='') as f:
            f.write(text)
        with open(path, newline='') as f:
            read = f.read()
        assert read == text
